
#include "lg/interface.h"
#include "lg/scrservices.h"
#include "ScriptLib.h"
#include "GenericTrigger.h"


#if (_DARKGAME == 3)
#define CDLINK	"SwitchLink"
#define TRAPFLAGS_PROPNAME	"TripFlags"
#else
#define CDLINK	"ControlDevice"
#define TRAPFLAGS_PROPNAME	"TrapFlags"
#endif

void cScrTrigger::InitTrigVars()
{
#if (_DARKGAME != 1)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, TRAPFLAGS_PROPNAME))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, TRAPFLAGS_PROPNAME, NULL);
		m_iFlags = static_cast<int>(mpTrapFlags);
	}
#if (_DARKGAME == 3)
	// Translate TripFlags to TrapFlags
	if (m_iFlags != 0)
	{
		int iFlags = (m_iFlags & 0x18) >> 3;
		iFlags |= (m_iFlags & 0x2) << 1;
		iFlags |= (m_iFlags & 0x1) << 3;
		m_iFlags = iFlags;
	}
#endif
#endif
}

bool cScrTrigger::IsLocked(sScrMsg* pMsg)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	return pLock->IsLocked(m_iObjId);
}

void cScrTrigger::SetLock(bool bLock)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	cMultiParm mpLock = static_cast<int>(bLock);
	pPropSrv->Set(m_iObjId, "Locked", mpLock);
}

long cScrTrigger::DoTurnOn(const cMultiParm& mpData)
{
	if (m_iFlags & kTrapFlagNoOn)
		return 0;
	if (m_iFlags & kTrapFlagOnce)
		SetLock(true);
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	return pLS->BroadcastOnAllLinks(m_iObjId, (m_iFlags & kTrapFlagInvert) ? "TurnOff" : "TurnOn", 
			pLTS->LinkKindNamed(CDLINK), mpData);
}

long cScrTrigger::DoTurnOff(const cMultiParm& mpData)
{
	if (m_iFlags & kTrapFlagNoOff)
		return 0;
	if (m_iFlags & kTrapFlagOnce)
		SetLock(true);
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	return pLS->BroadcastOnAllLinks(m_iObjId, (m_iFlags & kTrapFlagInvert) ? "TurnOn" : "TurnOff", 
			    pLTS->LinkKindNamed(CDLINK), mpData);
}
