
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <vector>

#include "lg/interface.h"
#include "lg/scrservices.h"
#include "lg/scrmanagers.h"
#include "lg/scrmsgs.h"
#include "lg/properties.h"
#include "lg/propdefs.h"
#include "lg/objects.h"
#include "lg/input.h"
#include "lg/quest.h"
#include "ScriptLib.h"
#include "tnhScript.h"

#if (_DARKGAME == 1)
#define SOUND_NET	
#else
#define SOUND_NET	, kSoundNetwork0
#endif

/*******************
 * ScriptController
 */
long cScr_ScriptCtrl::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	cMultiParm mpCtrl = const_cast<char*>(pszControl);
	int iDestObj = 0;
	if (pMsg->data2.type == kMT_String && pMsg->data2.psz != NULL)
	{
		iDestObj = StrToObject(pMsg->data2.psz);
	}
	if (iDestObj)
	{
		g_pScriptManager->SendMessage2(*static_cast<cMultiParm*>(pReply), m_iObjId, iDestObj, 
				"ScriptControl", mpCtrl, g_mpUndef, g_mpUndef);
	}
	else
	{
		CDSend("ScriptControl", m_iObjId, mpCtrl);
		if (pReply)
			MultiParmAssignInt(pReply, 1);
	}
	return 0;
}


/*******************
 * CommandControl
 */
long cScr_CommandCtrl::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
	char* pszCommand;
	for (pszCommand = strtok(const_cast<char*>(pszControl), ";"); pszCommand; pszCommand = strtok(NULL, ";"))
	{
		pDebugSrv->Command(pszCommand, "", "", "", "", "", "", "");
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapFrobber
 */
long cScr_TrapFrobber::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	sFrobMsg pFrobMsg;
	//sMultiParm mpRet;
	pFrobMsg.from = m_iObjId;
	pFrobMsg.to = m_iObjId;
	pFrobMsg.time = pMsg->time;
	pFrobMsg.flags = 4; // And someday I might even learn what this means.
	pFrobMsg.SrcObjId = m_iObjId;
	pFrobMsg.DstObjId = 0;
	pFrobMsg.Frobber = 0;
	if (GetObjectParamInt(m_iObjId, "player_frob"))
		pFrobMsg.Frobber = StrToObject("Player");
	pFrobMsg.DstLoc = kFrobLocNone;
	pFrobMsg.Sec = 0.0;
	pFrobMsg.Abort = 0;
	if (GetObjectParamInt(m_iObjId, "inv"))
	{
		pFrobMsg.message = "FrobInvEnd";
		pFrobMsg.SrcLoc = kFrobLocInv;
	}
	else
	{
		pFrobMsg.message = "FrobWorldEnd";
		pFrobMsg.SrcLoc = kFrobLocWorld;
	}

	g_pScriptManager->SendMessage(&pFrobMsg, pReply);
	return 0;
}


/*******************
 * TrigFrobber
 */
long __stdcall cScr_TrigFrobber::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (GetObjectParamInt(m_iObjId, "inv"))
		RegisterMessageHandler("FrobInvEnd", static_cast<MessageHandlerFunc>(&cScr_TrigFrobber::OnFrob));
	else
		RegisterMessageHandler("FrobWorldEnd", static_cast<MessageHandlerFunc>(&cScr_TrigFrobber::OnFrob));
	return 0;
}

long __stdcall cScr_TrigFrobber::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrMsg msg;
	msg.from = 0; // odd... the Send/PostMessage loses if I use m_iObjId.
	msg.to = m_iObjId;
	msg.time = pMsg->time;
	msg.flags = 8; // And someday I might even learn what this means.

	char* pszMsg = GetObjectParamString(m_iObjId, "frobmsg");
	msg.message = IF_NOT(pszMsg,"TurnOn");

	g_pScriptManager->SendMessage(&msg, pReply);
	if (pszMsg)
		g_pMalloc->Free(pszMsg);
	return 0;
}


/*******************
 * TrapMoveRelative
 */
long __stdcall cScr_MoveRelativeTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

/*
 * cScrVec.pos.x = Bank;
 * cScrVec.pos.y = Pitch;
 * cScrVec.pos.z = Heading;
 */
struct MoveRelative_Struct
{
	cScrVec	vPosition,
			vFacing,
			vVelocity,
			vRotation;
	bool bRotateAxes;
};

int cScr_MoveRelativeTrap::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	MoveRelative_Struct* pSDelta = reinterpret_cast<MoveRelative_Struct*>(pData);

	sLink sl;
	pLQ->Link(&sl);
	int iTargetObj = sl.dest;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	if (pSDelta->bRotateAxes)
	{
		pObjSrv->Teleport(iTargetObj, pSDelta->vPosition, pSDelta->vFacing, iTargetObj);
	}
	else
	{
		cScrVec vDestPos, vDestFacing;
		pObjSrv->Position(vDestPos,iTargetObj);
		vDestPos += pSDelta->vPosition;
		pObjSrv->Facing(vDestFacing,iTargetObj);
		vDestFacing += pSDelta->vFacing;
		pObjSrv->Teleport(iTargetObj, vDestPos, vDestFacing, 0);
	}

	/*
	SService<IPhysSrv> pPS(g_pScriptManager);
	pPS->GetVelocity(iTargetObj,vDestVel);
	vDestVel += pSDelta->vVelocity;
	pPS->SetVelocity(iTargetObj, vDestVel);
	*/
	SInterface<IPropertyManager> pPropMan(g_pScriptManager);
	SInterface<IProperty> pProp;
	pProp = pPropMan->GetPropertyNamed("PhysState");
	if (pProp)
	{
		IGenericProperty* pPhysStateProp;
		sPhysState* pPhysState;
		if (E_NOINTERFACE != pProp->QueryInterface(IID_IPhysStateProperty, reinterpret_cast<void**>(&pPhysStateProp)))
		{
			if (pPhysStateProp->Get(iTargetObj, reinterpret_cast<void**>(&pPhysState)))
			{
				cScrVec v;
				v = pPhysState->vels;
				pPhysState->vels = v + pSDelta->vVelocity;
				v = pPhysState->rot_vels;
				pPhysState->rot_vels = v + pSDelta->vRotation;
				pPhysStateProp->Set(iTargetObj, reinterpret_cast<void*>(pPhysState));
			}
			pPhysStateProp->Release();
		}
	}

	return 1;
}

long cScr_MoveRelativeTrap::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	MoveRelative_Struct vDelta;

	char* pszParams = new char[strlen(pszControl)+1];
	strcpy(pszParams, pszControl);
	char* pszCurrent;
	for (pszCurrent = strtok(pszParams, ";"); pszCurrent; pszCurrent = strtok(NULL, ";"))
	{
		// position/heading or velocity/rotation?
		bool delta;
		if (tolower(*pszCurrent) == 'd')
		{
			delta = TRUE;
			pszCurrent++;
		}
		else
			delta = FALSE;
		// get value
		char* pszValue = strchr(pszCurrent, '=') + 1;
		float value = strtod(pszValue, NULL);
		switch (tolower(*pszCurrent))
		{
		  case 'x':
			if (delta)
				vDelta.vVelocity.pos.x = value;
			else
				vDelta.vPosition.pos.x = value;
			break;
		  case 'y':
			if (delta)
				vDelta.vVelocity.pos.y = value;
			else
				vDelta.vPosition.pos.y = value;
			break;
		  case 'z':
			if (delta)
				vDelta.vVelocity.pos.z = value;
			else
				vDelta.vPosition.pos.z = value;
			break;
		  case 'b':
			if (delta)
				vDelta.vRotation.pos.x = value;
			else
				vDelta.vFacing.pos.x = value;
			break;
		  case 'p':
			if (delta)
				vDelta.vRotation.pos.y = value;
			else
				vDelta.vFacing.pos.y = value;
			break;
		  case 'h':
			if (delta)
				vDelta.vRotation.pos.z = value;
			else
				vDelta.vFacing.pos.z = value;
			break;
		}
	}
	delete[] pszParams;

	vDelta.bRotateAxes = GetObjectParamInt(m_iObjId, "rotate_axes");
	IterateLinks(
#if (_DARKGAME == 3)
		"SwitchLink",
#else
		"ControlDevice",
#endif
		m_iObjId, 0, LinkIterFunc, this, reinterpret_cast<void*>(&vDelta));

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


#if (_DARKGAME != 3)
/*******************
 * TrapCampaignQVar
 */
long __stdcall cScr_CampaignQVarTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	char* pszName = GetQVar(NULL, NULL);
	if (pszName)
	{
		SService<IQuestSrv> pQSrv(g_pScriptManager);
		if (pQSrv->Exists(pszName))
		{
			int iCurrent = pQSrv->Get(pszName);
			pQSrv->Delete(pszName);
			pQSrv->Set(pszName, iCurrent, (static_cast<sSimMsg*>(pMsg)->fStarting)?0:1);
		}
		delete[] pszName;
	}
	return 1;
}

char* cScr_CampaignQVarTrap::GetQVar(char* op, int* val)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (!pPropSrv->Possessed(m_iObjId,"TrapQVar"))
		return NULL;
	cMultiParm mpQVar;
	pPropSrv->Get(mpQVar, m_iObjId, "TrapQVar", NULL);
	if (mpQVar.type != kMT_String || static_cast<const char*>(mpQVar) == NULL)
		return NULL;
	
	char* pszProp = mpQVar.psz;
	if (op)
		*op = *pszProp;
	if (val)
		*val = strtol(pszProp+1, NULL, 10);
	char* pszName = strchr(pszProp, ':');
	if (!pszName)
		return NULL;
	char* pszDup = new char[strlen(++pszName)+1];
	strcpy(pszDup, pszName);
	mpQVar.Free();
	return pszDup;
}

long cScr_CampaignQVarTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	char cOperation;
	int iValue = 0;
	int iCurrent = 0;
	char *pszName = GetQVar(&cOperation, &iValue);
	if (!pszName)
		return 1;
	if (!bTurnOn)
		cOperation |= 0x80;		// flag high-bit for inverse operation
	SService<IQuestSrv> pQSrv(g_pScriptManager);
	if (pQSrv->Exists(pszName))
	{
		iCurrent = pQSrv->Get(pszName);
		//pQSrv->Delete(pszName);	// db management done in Sim now
	}
	switch (cOperation)
	{
	  case '=':
		iCurrent = iValue;
		break;
	  case '!':
	  case '|':
		iCurrent ^= iValue;
		break;
	  case ('!'+0x80):
	  case ('|'+0x80):
		iCurrent &= ~iValue;
		break;
	  case '+':
	  case ('-'+0x80):
		iCurrent += iValue;
		break;
	  case '-':
	  case ('+'+0x80):
		iCurrent -= iValue;
		break;
	  case '*':
	  case ('/'+0x80):
	  case ('%'+0x80):
		iCurrent *= iValue;
		break;
	  case '/':
	  case ('*'+0x80):
		iCurrent /= iValue;
		break;
	  case '%':
		iCurrent %= iValue;
		break;
	  case '{':
	  case ('}'+0x80):
		iCurrent <<= iValue;
		break;
	  case '}':
	  case ('{'+0x80):
		iCurrent >>= iValue;
		break;
	  case '\"':
		iCurrent = (iCurrent * 10) + (iValue % 10);
		break;
	  case '#':
		iCurrent = ((iCurrent * 10) + (iValue % 10)) % 10000;
		break;
	  case ('\"'+0x80):
	  case ('#'+0x80):
		iCurrent /= 10;
		break;
	  case '?':
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent += pDS->RandInt(0,iValue);
		break;
	  }
	  case ('?'+0x80):
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent -= pDS->RandInt(0,iValue);
		break;
	  }
	  case 'd':
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent += pDS->RandInt(1,iValue);
		break;
	  }
	  case ('d'+0x80):
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent -= pDS->RandInt(1,iValue);
		break;
	  }
	}
	pQSrv->Set(pszName, iCurrent, 0);
	delete[] pszName;
	return 0;
}
#endif


/*****************
 * TrapKillEvents
 */
long cScr_KillEvents::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SInterface<IInputBinder> pInput(g_pScriptManager);
	SService<IDarkInvSrv> pInvSrv(g_pScriptManager);
	if (bTurnOn)
	{
		long lEventMask = 0;
		char* pszParams = GetObjectParamString(m_iObjId, "event_mask");
		if (pszParams)
		{
			if (*pszParams == '$')
			{
				char* pszStart = pszParams + 1;
				char* pszEnd = pszStart;
				while (isalnum(*pszEnd) || (*pszEnd == '_'))
					++pszEnd;
				*pszEnd = '\0';
				if (pszEnd > pszStart)
				{
					SService<IQuestSrv> pQS(g_pScriptManager);
					if (pQS->Exists(pszStart))
					{
						lEventMask = pQS->Get(pszStart);
					}
				}
			}
			else
			{
				lEventMask = strtol(pszParams, NULL, 2);
			}
			g_pMalloc->Free(pszParams);
		}
		long lEventFlags = pInput->GetValidEvents();
		pInput->SetValidEvents(lEventFlags & lEventMask);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOff);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOff);

#if (_DARKGAME == 1)
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
		{
			cMultiParm mpTiming = 0;
			pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
			m_iTiming = IF_NOT(static_cast<int>(mpTiming),10000);
		}
		else
			m_iTiming = 10000;
#else
		if (m_iTiming == 0)
			m_iTiming = 10000;
#endif
	}
	else
	{
		pInput->SetValidEvents(0xFFFFFFFF);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOn);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOn);
	}
	return 0;
}


/****************
 * TrapCamAttach
 */
#if (_DARKGAME == 2)
long cScr_CamAttach::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (GetObjectParamInt(m_iObjId, "nolens"))
		{
			SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
			char sz[32];
			sprintf(sz, "cam_attach %d", m_iObjId);
			pDebugSrv->Command(sz, "", "", "", "", "", "", "");
		}
		else
		{
			SService<ICameraSrv> pCam(g_pScriptManager);
			if (GetObjectParamInt(m_iObjId, "static"))
				pCam->StaticAttach(m_iObjId);
			else
				pCam->DynamicAttach(m_iObjId);
		}
	}
	else
	{
		SService<ICameraSrv> pCam(g_pScriptManager);
		if (GetObjectParamInt(m_iObjId, "forcereturn"))
			pCam->ForceCameraReturn();
		else
			pCam->CameraReturn(m_iObjId);
	}
	return 0;
}
#elif (_DARKGAME == 3)
long cScr_CamAttach::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IShockGameSrv> pShock(g_pScriptManager);
	if (bTurnOn)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrStr strName;
		pOS->GetName(strName, m_iObjId);
		pShock->AttachCamera(strName);
		pShock->NoMove(1);
	}
	else
	{
		pShock->NoMove(0);
		pShock->AttachCamera("Player");
	}
	return 0;
}
#endif


/*******************
 * TrapFreezePlayer
 */
long cScr_FreezePlayer::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iIsFrozen == 0)
	{
		float fValue = GetObjectParamFloat(m_iObjId, "freeze_speed");
		char* pszName = GetObjectParamString(m_iObjId, "freeze_name");
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		pDISrv->AddSpeedControl(IF_NOT(pszName,"Freeze"), fValue, fValue);
		m_iIsFrozen = 1;
		if (pszName)
			g_pMalloc->Free(pszName);
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iIsFrozen)
	{
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		char* pszName = GetObjectParamString(m_iObjId, "freeze_name");
		if (pszName)
		{
			pDISrv->RemoveSpeedControl(pszName);
			g_pMalloc->Free(pszName);
		}
		else
			pDISrv->RemoveSpeedControl("Freeze");
		m_iIsFrozen = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkInvSrv> pDISrv(g_pScriptManager);
	char* pszName = GetObjectParamString(m_iObjId, "freeze_name");
	if (m_iIsFrozen)
		pDISrv->RemoveSpeedControl(IF_NOT(pszName,"Freeze"));
	pDISrv->AddSpeedControl(IF_NOT(pszName,"Freeze"), fValue, fValue);
	m_iIsFrozen = 1;
	if (pszName)
		g_pMalloc->Free(pszName);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapThreshold
 */
long __stdcall cScr_ThresholdRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iGlobalCounter.Init(GetObjectParamInt(m_iObjId, "initial"));
	return 0;
}

long cScr_ThresholdRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iThreshold = GetObjectParamInt(m_iObjId, "threshold");
	if (GetObjectParamInt(m_iObjId, "global") || pMsg->from == 0)
	{
		if (++m_iGlobalCounter == iThreshold)
		{
			m_iGlobalCounter = 0;
			CDSend("TurnOn", m_iObjId);
			return 0;
		}
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = GetObjectParamString(m_iObjId, "initial");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = 0;
		if (mpCounter.psz)
		{
			iCounter = strtol(static_cast<const char*>(mpCounter), NULL, 10);
			mpCounter.Free();
		}
		if (++iCounter == iThreshold)
		{
			CDSend("TurnOn", m_iObjId);
			mpCounter = "0";
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
			return 0;
		}
		char psz[16];
		sprintf(psz, "%d", iCounter);
		mpCounter = psz;
		pLTS->LinkSetData(lCounter, NULL, mpCounter);
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}

long cScr_ThresholdRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iFloor = GetObjectParamInt(m_iObjId, "floor");
	if (GetObjectParamInt(m_iObjId, "global") || pMsg->from == 0)
	{
		if (m_iGlobalCounter > iFloor)
			--m_iGlobalCounter;
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = GetObjectParamString(m_iObjId, "floor");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = 0;
		if (mpCounter.psz)
		{
			iCounter = strtol(static_cast<const char*>(mpCounter), NULL, 10);
			mpCounter.Free();
		}
		if (iCounter > iFloor)
		{
			char psz[16];
			sprintf(psz, "%d", iCounter - 1);
			mpCounter = psz;
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}


/*******************
 * TrapRandomTimer
 */
// Relay to all CD links.
long __stdcall cScr_RandomTimer::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "TurnOn") && 
		stricmp(pszMsgName, "TurnOff"))
		return 1;
	
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	RelayMessage(pszMsgName);
	return 0;
}

// Delay for a random amount of time.
long __stdcall cScr_RandomTimer::OnTurnOnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (!pLock->IsLocked(m_iObjId))
	{
		unsigned long lFlags = 0;
#if (_DARKGAME == 2)
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
		{
			cMultiParm mpFlags;
			pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
			lFlags = static_cast<int>(mpFlags);
		}
#elif (_DARKGAME == 3)
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
		{
			cMultiParm mpFlags;
			pPropSrv->Get(mpFlags, m_iObjId, "TripFlags", NULL);
			lFlags = static_cast<int>(mpFlags);
			lFlags = ((lFlags & 0x18) >> 3) | ((lFlags & 0x2) << 1) | ((lFlags & 0x1) << 3);
		}
#endif
		const char* pszMsgName = pMsg->message;
		if ((lFlags & kTrapFlagNoOn) && !stricmp(pszMsgName, "TurnOn"))
			return 1;
		if ((lFlags & kTrapFlagNoOff) && !stricmp(pszMsgName, "TurnOff"))
			return 1;
		if (lFlags & kTrapFlagInvert)
		{
			if (!stricmp(pszMsgName, "TurnOn"))
				pszMsgName = "TurnOff";
			else
				pszMsgName = "TurnOn";
		}

		int iMinTime = 0, iMaxTime = 0;
		char* pszParamString = GetObjectParams(m_iObjId);
		if (pszParamString)
		{
			iMinTime = GetParamTime(pszParamString, "min_time");
			iMaxTime = GetParamTime(pszParamString, "max_time");
			g_pMalloc->Free(pszParamString);
		}

		unsigned long iInterval;
		if (iMinTime >= iMaxTime)
			iInterval = iMaxTime;
		else
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			iInterval = pDS->RandInt(iMinTime, iMaxTime);
		}
		if (iInterval > 0)
			g_pScriptManager->SetTimedMessage2(m_iObjId, pszMsgName, iInterval, kSTM_OneShot, g_mpUndef);
		else
			RelayMessage(pszMsgName);
	}
	return 0;
}

void cScr_RandomTimer::RelayMessage(const char* pszMsgName)
{
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TripFlags", NULL);
		if (static_cast<int>(mpFlags) & 0x8)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}
#endif
	CDSend(pszMsgName, m_iObjId);
}


/***********************
 * TrapRepeatingRelay
 */
long __stdcall cScr_RepeatingRelay::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		if (GetObjectParamInt(m_iObjId, "autostart"))
			OnTurnOn(pMsg, pReply, eTrace);
	}
	return 1;
}

// Send TurnOn to all CD links.
// Then decrement the param and, if zero, stop the timer.
long __stdcall cScr_RepeatingRelay::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "Repeat"))
		return 1;
	
	CDSend("TurnOn", m_iObjId);

	int iCount = m_iRepeatCount;
	if (iCount > 0)
	{
		if (--iCount == 0)
		{
			if (m_hTimerHandle != NULL)
			{
				g_pScriptManager->KillTimedMessage(m_hTimerHandle);
				m_hTimerHandle = NULL;
			}
		}
		else
			m_iRepeatCount = iCount;
	}
	
	return 0;
}

long __stdcall cScr_RepeatingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hTimerHandle == NULL)
		return 1;
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (!(static_cast<int>(mpFlags) & kTrapFlagNoOff))
		{
			g_pScriptManager->KillTimedMessage(m_hTimerHandle);
			m_hTimerHandle = NULL;
		}
	}
	else
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TripFlags", NULL);
		if (!(static_cast<int>(mpFlags) & 0x1))
		{
			g_pScriptManager->KillTimedMessage(m_hTimerHandle);
			m_hTimerHandle = NULL;
		}
	}
	else
#endif
	{
		g_pScriptManager->KillTimedMessage(m_hTimerHandle);
		m_hTimerHandle = NULL;
	}
	return 0;
}

// Send the message, then set a timer.
// Params are "repeat" and "interval".
long __stdcall cScr_RepeatingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	int iCount = -1;
	int iInterval = 0;
	char* pszParamString = GetObjectParams(m_iObjId);
	if (pszParamString)
	{
		iCount = GetParamInt(pszParamString, "repeat");
		iInterval = GetParamTime(pszParamString, "interval");
		g_pMalloc->Free(pszParamString);
	}

	if (iInterval <= 0)
		return 1;

	if (iCount > 0)
	{
		if (m_hTimerHandle == NULL)
		{
			m_iRepeatCount = --iCount;
			CDSend("TurnOn", m_iObjId);
			if (iCount)
			{
				m_hTimerHandle = g_pScriptManager->SetTimedMessage2(m_iObjId, "Repeat", iInterval, kSTM_Periodic, g_mpUndef);
			}
		}
		else // just prolong the existing timer
			m_iRepeatCount += iCount;
	}
	else if (iCount < 0)
	{
		if (m_hTimerHandle == NULL)
		{
			m_iRepeatCount = -1;
			CDSend("TurnOn", m_iObjId);
			m_hTimerHandle = g_pScriptManager->SetTimedMessage2(m_iObjId, "Repeat", iInterval, kSTM_Periodic, g_mpUndef);
		}
	}

#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TripFlags", NULL);
		if (static_cast<int>(mpFlags) & 0x8)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "Locked", mpFlags);
		}
	}
#endif

	return 0;
}


#if (_DARKGAME == 2)
/*********************
 * TrapRenderFlash
 */
int cScr_RenderFlash::Relink(ILinkSrv* pLS, ILinkToolsSrv* pLTS, long flavor, int source, int dest)
{
	link l;
	pLS->GetOne(l, flavor, source, 0);
	int ret = 0;
	if (l)
	{
		sLink sl;
		pLTS->LinkGet(l, sl);
		ret = sl.dest;
		pLS->Destroy(l);
	}
	link nl;
	// RenderFlash links don't have any data, so no worries.
	pLS->Create(nl, flavor, source, dest);
	return ret;
}

long cScr_RenderFlash::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (!bTurnOn)
		return 1;
	
	object oPlayer = StrToObject("Player");
	// A bit of trickery here.  The renderflash is linked from the archetype of "Player",
	// not the concrete object. 
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	int iAvatar = pTraits->GetArchetype(oPlayer);
	if (iAvatar != 0 && oPlayer != 0)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		// The render flash that is linked from "Garrett" will be shown.
		long flavor = pLTS->LinkKindNamed("RenderFlash");
		long l = 0;
		sLink sl;
		l = GetAnyLinkInheritedSrc("RenderFlash", m_iObjId, 0, &sl);
		if (l != 0)
		{
			//pLTS->LinkGet(l, sl);
			int oldflash = Relink(pLS, pLTS, flavor, iAvatar, sl.dest);
			SService<ICameraSrv> pCam(g_pScriptManager);
			SService<ISoundScrSrv> pSound(g_pScriptManager);
			true_bool ret;
			int iSchema = GetAnyLinkDest("SoundDescription", m_iObjId);
			if (iSchema)
				pSound->PlaySchemaAmbient(ret, oPlayer, iSchema, kSoundNetwork0);
			pCam->StaticAttach(oPlayer);
			pCam->CameraReturn(oPlayer);
			Relink(pLS, pLTS, flavor, iAvatar, oldflash);
		}
	}
	return 0;
}
#endif


#if (_DARKGAME != 3)
/*******************
 * TrapDisarmPlayer
 */
int cScr_DisarmPlayer::DisarmIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	IObjectSrv* pOS = reinterpret_cast<IObjectSrv*>(pData);
	sLink sl;
	pLQ->Link(&sl);
	int iObjId = static_cast<cScr_DisarmPlayer*>(pScript)->m_iObjId;
	bool bPotions = true;
	char* pszParam = GetObjectParamString(iObjId, "nopotions");
	if (pszParam)
	{
		bPotions = strtol(pszParam, NULL, 10) == 0;
		g_pMalloc->Free(pszParam);
	}
	object oWeapons, oProjectiles, oCrystals, oGrenadz, oPotions;
	pOS->Named(oWeapons, "Weapon");
	pOS->Named(oProjectiles, "Projectile");
	pOS->Named(oCrystals, "Crystal");
	pOS->Named(oGrenadz, "Grenadz");
	pOS->Named(oPotions, "Potions");
	true_bool child_p;
	if (*(pOS->InheritsFrom(child_p, sl.dest, oWeapons)) ||
		*(pOS->InheritsFrom(child_p, sl.dest, oProjectiles)) ||
		*(pOS->InheritsFrom(child_p, sl.dest, oCrystals)) ||
		*(pOS->InheritsFrom(child_p, sl.dest, oGrenadz)) ||
		(bPotions && *(pOS->InheritsFrom(child_p, sl.dest, oPotions))) )
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->Remove(sl.dest, 0);
		pContainSrv->Add(sl.dest, iObjId, 0, 1);
	}
	return 1;
}

long cScr_DisarmPlayer::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oPlayer;
	pObjSrv->Named(oPlayer, "Player");
	if (bTurnOn)
	{
		IterateLinks("Contains", oPlayer, 0, DisarmIterFunc, this, reinterpret_cast<void*>(pObjSrv.get()));
	}
	else
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->MoveAllContents(m_iObjId, oPlayer, 1);
	}
	return 0;
}
#endif


#if (_DARKGAME != 3)
/*******************
 * ObjFrobInert
 */
int cScr_BaseFrobInert::MakeUnfrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	AddSingleMetaProperty("FrobInert", sl.dest);
	return 1;
}

int cScr_BaseFrobInert::MakeFrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	RemoveSingleMetaProperty("FrobInert", sl.dest);
	return 1;
}

void cScr_BaseFrobInert::DoFrobInert(bool bFrobbable)
{
	IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, 
			(bFrobbable) ? MakeFrobbableIterFunc : MakeUnfrobbableIterFunc, this, NULL);
}

bool cScr_BaseFrobInert::IsFrobbable(void)
{
	return m_iDead != 0;
}

// Add or Remove the metaproperty based on the current door state.
long __stdcall cScr_BaseFrobInert::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		m_iDead.Init(0);
		DoFrobInert(IsFrobbable());
	}	
	return 1;
}

long __stdcall cScr_BaseFrobInert::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iDead = 1;
	DoFrobInert(true);
	return 0;
}

// Force frobbable.
long __stdcall cScr_BaseFrobInert::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoFrobInert(true);
	if (m_bSim)
		m_iDead = 1;
	return 0;
}



/*******************
 * DoorFrobInert
 */
bool cScr_DoorFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0)
		return true;
	SService<IDoorSrv> pDoor(g_pScriptManager);
	int iState = pDoor->GetDoorState(m_iObjId);
	return iState != kDoorStateClosed;
}

// Door is dead. We may or may not want to terminally remove the metaproperty.
// Check the "breakable" parameter to see if we should treat the door
// as permanently open.
// If a door is destroyed without getting Slain first, there will likely be an
// ugly black rectangle where the door was.
long __stdcall cScr_DoorFrobInert::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (GetObjectParamInt(m_iObjId, "breakable"))
	{
		m_iDead = 1;
		// StdDoor will open the door for us.
	}
	return 0;
}

// Remove the metaproperty.
long __stdcall cScr_DoorFrobInert::OnDoorOpening(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoFrobInert(true);
	return 0;
}

// Add the metaproperty (if it doesn't already exist).
long __stdcall cScr_DoorFrobInert::OnDoorClose(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_iDead)
		DoFrobInert(false);
	return 0;
}


/*******************
 * GrabFrobInert
 */
long __stdcall cScr_GrabFrobInert::OnFrobWorldEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iDead = 1;
	DoFrobInert(true);
	return 0;
}


/*******************
 * TweqFrobInert
 */
bool cScr_TweqFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0)
		return true;
	SService<IPropertySrv> pProps(g_pScriptManager);
	if (!pProps->Possessed(m_iObjId, "StTweqJoints"))
		return true;
	
	cMultiParm mpJoint;
	pProps->Get(mpJoint, m_iObjId, "CfgTweqJoints", "Primary Joint");
	int iJoint = static_cast<int>(mpJoint);
	if (iJoint == 0)
		return true;
	char szAnims[12];
	snprintf(szAnims, 12, "Joint%dAnimS", iJoint);
	pProps->Get(mpJoint, m_iObjId, "StTweqJoints", szAnims);
	return (static_cast<int>(mpJoint) & kTweqFlagReverse) != 0;
}

long __stdcall cScr_TweqFrobInert::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	if (pTweqMsg->Op == kTweqOpHaltTweq)
	{
		DoFrobInert(pTweqMsg->Dir == 0);
	}
	return 0;
}


/*******************
 * TrapFrobInert
 */
bool cScr_TrapFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0 || m_iFrobbable != 0)
		return true;
	return false;
}

long __stdcall cScr_TrapFrobInert::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

#if (_DARKGAME == 2)
	int iFlags = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}
	if (iFlags & kTrapFlagNoOn)
		return 1;
	
	if (iFlags & kTrapFlagInvert)
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
	}
	else
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
	}

	int iTiming = 0;
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming = 0;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		iTiming = static_cast<int>(mpTiming);
	}
	if (iTiming > 0)
	{
		g_pScriptManager->SetTimedMessage2(m_iObjId, (iFlags & kTrapFlagInvert) ? "TurnOn" : "TurnOff",
				iTiming, kSTM_OneShot, 0);
	}

	if (iFlags & kTrapFlagOnce)
	{
		cMultiParm mpLock = true;
		pPropSrv->Set(m_iObjId, "Locked", mpLock);
	}
#else
	DoFrobInert(true);
	m_iFrobbable = 1;
#endif
	return 0;
}

long __stdcall cScr_TrapFrobInert::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

#if (_DARKGAME == 2)
	int iFlags = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}
	if (iFlags & kTrapFlagNoOff)
		return 1;
	
	if (iFlags & kTrapFlagInvert)
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
	}
	else
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
	}

	if (iFlags & kTrapFlagOnce)
	{
		cMultiParm mpLock = true;
		pPropSrv->Set(m_iObjId, "Locked", mpLock);
	}
#else
	DoFrobInert(false);
	m_iFrobbable = 0;
#endif
	return 0;
}

long __stdcall cScr_TrapFrobInert::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	if (!stricmp(static_cast<const char*>(pTimerMsg->name), "TurnOn"))
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
		return 0;
	}
	else if (!stricmp(static_cast<const char*>(pTimerMsg->name), "TurnOff"))
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
		return 0;
	}
	return 1;
}
#endif


/********************
 * UnlinkOnContained
 */
// Remove links from a specified object.
// Only happens once to save time.
long __stdcall cScr_ContainedUnlink::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_bSim || m_iAlreadyUnlinked != 0)
		return 0;
	if (static_cast<sContainedScrMsg*>(pMsg)->event != kContainAdd && 
		static_cast<sContainedScrMsg*>(pMsg)->event != kContainCombine)
		return 1;

#if (_DARKGAME == 3)
	SService<INetworkingSrv> pNet(g_pScriptManager);
	if (pNet->IsPlayer(static_cast<sContainedScrMsg*>(pMsg)->container))
#else
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oAvatar;
	pObjSrv->Named(oAvatar, "Avatar");
	true_bool bIsAvatar;
	if (*(pObjSrv->InheritsFrom(bIsAvatar, static_cast<sContainedScrMsg*>(pMsg)->container, oAvatar)))
#endif
	{
		IterateLinksByData("ScriptParams", m_iObjId, 0, "Unlink", -1, ScriptParamsIterFunc, this, NULL);
		m_iAlreadyUnlinked = 1;
		return 0;
	}
	return 1;
}

int cScr_ContainedUnlink::ScriptParamsIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	IterateLinks(NULL, sl.dest, sl.source, DestroyLinksIterFunc, pScript, NULL);
	/* Still haven't worked out how to use DestroyMany
	char szFrom[12];
	char szTo[12];
	sprintf(szFrom, "%d", sl.dest);
	sprintf(szTo, "%d", sl.source);
	pLS->DestroyMany(0, szFrom, szTo);
	*/
	pLS->Destroy(pLQ->ID());
	return 1;
}

int cScr_ContainedUnlink::DestroyLinksIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	pLS->Destroy(pLQ->ID());
	return 1;
}


#if (_DARKGAME != 3)
/******************
 * TimedGrenade
 */
long __stdcall cScr_TimedGrenade::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iSafety.Init();
	m_hFuseTimer.Init();

	/* Bad idea. Use a receptron link to send to script
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	SService<IActReactSrv> pARSrv(g_pScriptManager);
	object oStim;
	pObjSrv->Named(oStim, "FireStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "BashStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "PokeStim");
	pARSrv->SubscribeToStimulus(m_iObjId, oStim);
	*/
	if (m_hFuseTimer)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	return 0;
}

long __stdcall cScr_TimedGrenade::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	/*
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	SService<IActReactSrv> pARSrv(g_pScriptManager);
	object oStim;
	pObjSrv->Named(oStim, "FireStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "BashStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	pObjSrv->Named(oStim, "PokeStim");
	pARSrv->UnsubscribeToStimulus(m_iObjId, oStim);
	*/
	return 1;
}

long __stdcall cScr_TimedGrenade::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);

	if (!stricmp(pszMsgName, "StackHack"))
	{
		RemoveObjectParam(m_iObjId, "launch_active");
		return 0;
	}
	if (!stricmp(pszMsgName, "GrenadeActive"))
	{
		GrenadeActive(true);
		return 0;
	}
	if (stricmp(pszMsgName, "TimedGrenade"))
		return 1;

	if (m_iSafety)
	{
		m_iSafety = 0;
		m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", 1600, kSTM_OneShot, g_mpUndef);
	}
	else
	{
		m_hFuseTimer = NULL;
		SService<IDamageSrv> pDS(g_pScriptManager);
		pDS->Slay(m_iObjId, 0);
	}
	return 0;
}

long __stdcall cScr_TimedGrenade::OnStimulus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Intensity is handled by the receptron link.
	//if (static_cast<sStimMsg*>(pMsg)->intensity < 5.0)
	//	return 1;
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	sLink lSource;
	pLTS->LinkGet(static_cast<sStimMsg*>(pMsg)->source, lSource);
	SService<IDamageSrv> pDS(g_pScriptManager);
	pDS->Slay(m_iObjId, lSource.source);
	return 0;
}

long __stdcall cScr_TimedGrenade::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
	
	GrenadeFlash();

	return 0;
}

long __stdcall cScr_TimedGrenade::OnFrobInv(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StackCount"))
	{
		cMultiParm mpStack;
		pPropSrv->Get(mpStack, m_iObjId, "StackCount", NULL);
		if (static_cast<int>(mpStack) > 1)
		{
			// This is gonna get ugly.
			char* pszParams = GetObjectParams(m_iObjId);
			char* pszNew;
			if (pszParams)
			{
				pszNew = new char[strlen(pszParams) + 17];
				sprintf(pszNew, "%s;%s", "launch_active=1", pszParams);
			}
			else
			{
				pszNew = new char[16];
				strcpy(pszNew, "launch_active=1");
			}
			SetObjectParams(m_iObjId, pszNew);
			delete[] pszNew;
			g_pScriptManager->SetTimedMessage2(m_iObjId, "StackHack", 1, kSTM_OneShot, g_mpUndef);
		}
		else
		{
			GrenadeArmed();
		}
	}
	else
	{
		GrenadeArmed();
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_TimedGrenade::OnFrobWorld(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
		if (m_iSafety)
		{
			m_iSafety = 0;
			GrenadeActive(false);
		}
		else
		{
			SService<IDamageSrv> pDS(g_pScriptManager);
			pDS->Slay(m_iObjId, static_cast<sFrobMsg*>(pMsg)->Frobber);
		}
		return 0;
	}
	// Just in case
	return 1;
}

long __stdcall cScr_TimedGrenade::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_bSim)
		return 0;
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	// Activate the schema after being removed from the inventory.
	if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainRemove && m_hFuseTimer)
	{
		GrenadeActive(true);
		return 0;
	}
	else if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainAdd
		  || static_cast<sContainedScrMsg*>(pMsg)->event == kContainCombine)
	{
		RemoveObjectParam(m_iObjId, "launch_active");
	}
	return 1;
}

long __stdcall cScr_TimedGrenade::OnCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Dynamically created objects don't get a Sim message. Use the message time instead.
	if (pMsg->time > 0 && GetObjectParamInt(m_iObjId, "launch_active"))
	{
		GrenadeArmed();
		g_pScriptManager->SetTimedMessage2(m_iObjId, "GrenadeActive", 1, kSTM_OneShot, g_mpUndef);
		RemoveObjectParam(m_iObjId, "launch_active");
		return 0;
	}
	return 1;
}

long __stdcall cScr_TimedGrenade::OnSchemaDone(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer) // We're active, so let's play the active loop.
	{
		SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
		true_bool bRet;
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj SOUND_NET);
		return 0;
	}
	// If not active... who cares?
	return 1;
}

void cScr_TimedGrenade::GrenadeArmed(void)
{
	if (m_hFuseTimer)
	{
		// How in the hell did you get here? Well, let's clean this up.
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	int iFuseTime = 5000;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		iFuseTime = mpTiming;
	}
	if (iFuseTime > 2000)
	{
		m_iSafety = 1;
		iFuseTime -= 1600;
	}
	else // Uh-oh...
	{
		m_iSafety = 0;
	}
	m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", iFuseTime, kSTM_OneShot, g_mpUndef);
}

void cScr_TimedGrenade::GrenadeFlash(void)
{
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	// TriggerWorldFlash ignores the RenderFlash link on the concrete object
	// ... the bastard
	int iObject = pTraits->GetArchetype(m_iObjId);
	if (GetOneLinkInheritedSrc("RenderFlash", iObject, 0, NULL) != 0)
	{
		SService<IDarkPowerupsSrv> pDPS(g_pScriptManager);
		pDPS->TriggerWorldFlash(m_iObjId);
	}
}

void cScr_TimedGrenade::GrenadeActive(bool bActivate)
{

	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	if (bActivate)
	{
		//pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Activate", m_iObjId, 0, kEnvSoundOnObj SOUND_NET);
		if (!bRet)
		{
			pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj SOUND_NET);
		}
	}
	else
	{
		pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Deactivate", m_iObjId, 0, kEnvSoundOnObj SOUND_NET);
	}

	SService<IActReactSrv> pARSrv(g_pScriptManager);
	cMultiParm& mpNull = const_cast<cMultiParm&>(g_mpUndef);
	cMultiParm mpType = kTweqTypeModels;
	cMultiParm mpAction = (bActivate) ? kTweqDoActivate : kTweqDoHalt;
	pARSrv->React(pARSrv->GetReactionNamed("tweq_control"), 1.0, m_iObjId, 0,
			  mpType, mpAction, mpNull, mpNull, mpNull, mpNull, mpNull, mpNull);
	if (!bActivate)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "CfgTweqModels"))
		{
			cMultiParm mpModel;
			pPropSrv->Get(mpModel, m_iObjId, "CfgTweqModels", "Model 5");
			if (mpModel.type == kMT_String && mpModel.psz != NULL)
			{
				pPropSrv->Set(m_iObjId, "ModelName", mpModel);
				mpModel.Free();
			}
		}
	}

	if (bActivate)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	else
	{
		UnregisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		UnregisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
}
#endif


#if (_DARKGAME != 3)
/****************
 * ZeroGravRoom
 *
 * This script was reverse-engineered from System Shock 2.
 * (In particular, I never would've known to set the base friction.)
 */
long __stdcall cScr_ZeroGravRoom::OnEnter(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", 0.0);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Base Friction", 1.0);

	return 0;
}

long __stdcall cScr_ZeroGravRoom::OnExit(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	float fGrav = 100.0;
	if (pPSrv->Possessed(pRMsg->ToObjId, "RoomGrav"))
	{
		cMultiParm mpGrav;
		pPSrv->Get(mpGrav, pRMsg->ToObjId, "RoomGrav", NULL);
		fGrav = static_cast<int>(mpGrav);
	}
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", fGrav);

	return 0;
}
#endif


/***************
 * TrapFadeOut
 */
#if (_DARKGAME == 3)
long __stdcall cScr_FadeOut::OnNetFade(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	float fValue = static_cast<float>(pMsg->data);
	int r = 0;
	int g = 0;
	int b = 0;
	char* pszParams = GetObjectParams(m_iObjId);
	if (pszParams)
	{
		r = GetParamInt(pszParams, "r");
		g = GetParamInt(pszParams, "g");
		b = GetParamInt(pszParams, "b");
		g_pMalloc->Free(pszParams);
	}

	SService<IShockGameSrv> pShock(g_pScriptManager);
	if (fValue < 0)
		pShock->StartFadeIn(int(-fValue * 1000.0), r,g,b);
	else
		pShock->StartFadeOut(int(fValue * 1000.0), r,g,b);

	return 0;
}
#endif

long cScr_FadeOut::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
#if (_DARKGAME == 3)
	SService<INetworkingSrv> pNet(g_pScriptManager);
	pNet->Broadcast(m_iObjId, "NetFade", 0, fValue);

	int r = 0;
	int g = 0;
	int b = 0;
	char* pszParams = GetObjectParams(m_iObjId);
	if (pszParams)
	{
		r = GetParamInt(pszParams, "r");
		g = GetParamInt(pszParams, "g");
		b = GetParamInt(pszParams, "b");
		g_pMalloc->Free(pszParams);
	}

	SService<IShockGameSrv> pShock(g_pScriptManager);
	if (fValue < 0)
		pShock->StartFadeIn(int(-fValue * 1000.0), r,g,b);
	else
		pShock->StartFadeOut(int(fValue * 1000.0), r,g,b);
#else
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(fValue);
#endif
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

#if (_DARKGAME != 3)
long cScr_FadeOut::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(-1);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}
#endif


#if (_DARKGAME != 3)
/*****************
 * LockHelper
 */
long __stdcall cScr_LockHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	if (pLockSrv->IsLocked(m_iObjId))
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobLock", -1, LinkIterFunc, this, reinterpret_cast<void*>(1));
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_LockHelper::OnMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// ILinkSrv::BroadcastOnAllLinks will do this, but it uses a case-sensitive compare.
	// I don't want to be that strict.
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	IterateLinksByData("ScriptParams", m_iObjId, 0, pMsg->message, -1, LinkIterFunc,
			    this, reinterpret_cast<void*>(pLockSrv->IsLocked(m_iObjId)));
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

int cScr_LockHelper::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iLocked = reinterpret_cast<int>(pData);
	sLink sl;
	pLQ->Link(&sl);
	SimplePost(sl.source, sl.dest, "TurnOn", iLocked);
	return 1;
}


/*****************
 * KeyHelper
 */
long __stdcall cScr_KeyHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "UseKey");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*****************
 * PickHelper
 */
long __stdcall cScr_PickHelper::OnFrobInvBegin(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickBegin");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_PickHelper::OnFrobInvEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickEnd");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}
#endif


/******************
 * TrapSiblingRelay
 */
long __stdcall cScr_SiblingRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iInit = 0;
	int iLowestSibling = FindFirstObject(m_iObjId);
	m_iBaseObjectId.Init(0);
	m_iCurrentState.Init(GetObjectParamInt(m_iObjId, "start_on"));
	m_iSwallowTurnOn.Init(0);
	m_iSwallowTurnOff.Init(0);
	if (m_iBaseObjectId == 0)
	{
		m_iBaseObjectId = iLowestSibling;
	}
	else if (m_iBaseObjectId != iLowestSibling)
	{
		// Joy. The base object ID has changed while we weren't looking.
		if (iLowestSibling == m_iObjId)
		{
			sScrDatumTag oldLock = { m_iBaseObjectId,"cScr_SiblingRelay","m_giSiblingLock"};
			if (g_pScriptManager->IsScriptDataSet(&oldLock))
			{
				sMultiParm param;
				param.type = kMT_Undef;
				g_pScriptManager->ClearScriptData(&oldLock, &param);
				iInit = param.i;
			}
		}
		m_iBaseObjectId = iLowestSibling;
	}
	m_giSiblingLock.SetObj(iLowestSibling);
	if (iLowestSibling == m_iObjId)
		m_giSiblingLock.Init(iInit);

	return 0;
}

long __stdcall cScr_SiblingRelay::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId == m_iObjId)
	{
		int iNewBase = FindFirstObject(0);
		if (iNewBase != 0)
		{
			int iSiblingLock = m_giSiblingLock;
			m_giSiblingLock.SetObj(iNewBase);
			m_giSiblingLock = iSiblingLock;
			SService<IPropertySrv> pPropSrv(g_pScriptManager);
			if (pPropSrv->Possessed(m_iObjId, "Locked"))
				pPropSrv->CopyFrom(iNewBase, "Locked", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "Locked");
#if (_DARKGAME == 2)
			if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
				pPropSrv->CopyFrom(iNewBase, "TrapFlags", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "TrapFlags");
			if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
				pPropSrv->CopyFrom(iNewBase, "ScriptTiming", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "ScriptTiming");
#elif (_DARKGAME == 3)
			if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
				pPropSrv->CopyFrom(iNewBase, "TripFlags", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "TripFlags");
			if (pPropSrv->Possessed(m_iObjId, "DelayTime"))
				pPropSrv->CopyFrom(iNewBase, "DelayTime", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "DelayTime");
#endif
			PropogateMessage("NewBaseObject", iNewBase);
		}
	}
	// Let's go ahead and re-link outside of sim-time.
	// Makes maintenance of networks in DromEd easier.
	int iLeft = 0, iRight = 0;
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(0, m_iObjId);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iLeft = sl.source;
			break;
		}
	}
	pLQ->Release();
	pLQ = pRel->Query(m_iObjId, 0);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iRight = sl.dest;
			break;
		}
	}
	pLQ->Release();
	if (iLeft && iRight)
		pRel->AddFull(iLeft, iRight, "Sibling");

	return 0;
}

long __stdcall cScr_SiblingRelay::OnNewBaseObject(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pMsg->data.type == kMT_Int && pMsg->data.i != 0)
	{
		m_iBaseObjectId = pMsg->data.i;
		m_giSiblingLock.SetObj(pMsg->data.i);
	}
	return 0;
}

long __stdcall cScr_SiblingRelay::OnNowLocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpLock = 1;
		pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
	}
	return 0;
}

long __stdcall cScr_SiblingRelay::OnNowUnlocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpLock = 0;
		pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
	}
	return 0;
}

void cScr_SiblingRelay::InitTrapVars()
{
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	int iObj = m_iBaseObjectId;
	if (pPropSrv->Possessed(iObj, "ScriptTiming"))
	{
		cMultiParm mpTiming = 0;
		pPropSrv->Get(mpTiming, iObj, "ScriptTiming", NULL);
		m_iTiming = static_cast<int>(mpTiming);
	}
	if (pPropSrv->Possessed(iObj, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, iObj, "TrapFlags", NULL);
		m_iFlags = static_cast<int>(mpTrapFlags);
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	int iObj = m_iBaseObjectId;
	if (pPropSrv->Possessed(iObj, "DelayTime"))
	{
		cMultiParm mpTiming = 0;
		pPropSrv->Get(mpTiming, iObj, "DelayTime", NULL);
		m_iTiming = int(static_cast<float>(mpTiming) * 1000.0);
	}
	if (pPropSrv->Possessed(iObj, "TripFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, iObj, "TripFlags", NULL);
		int iFlags = static_cast<int>(mpTrapFlags);
		m_iFlags = ((iFlags & 0x18) >> 3) | ((iFlags & 0x2) << 1) | ((iFlags & 0x1) << 3);
	}
#endif
}

// As well as reading from the base object, we always want to allow propogation msgs.
bool cScr_SiblingRelay::IsLocked(sScrMsg* pMsg)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	return m_giSiblingLock == 0 && pLock->IsLocked(static_cast<int>(m_iBaseObjectId));
}

void cScr_SiblingRelay::SetLock(bool bLock)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	cMultiParm mpLock = static_cast<int>(bLock);
	pPropSrv->Set(static_cast<int>(m_iBaseObjectId), "Locked", mpLock);
}

long cScr_SiblingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOn > 0)
	{
		--m_iSwallowTurnOn;
	}
	else
	{
		CDSend("TurnOn", m_iObjId);
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			char* pszParam = GetObjectParamString(m_iObjId, "exclusive");
			if (pszParam)
			{
				if (!stricmp(pszParam, "on") || !stricmp(pszParam, "all"))
					PropogateMessage(pszOff, 1);
				else
					PropogateMessage(pszOn, 1);
				g_pMalloc->Free(pszParam);
			}
			else
				PropogateMessage(pszOn, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(1);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 1)
				ToggleLevers(1);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 1;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_SiblingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOff > 0)
	{
		--m_iSwallowTurnOff;
	}
	else
	{
		CDSend("TurnOff", m_iObjId);
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			char* pszParam = GetObjectParamString(m_iObjId, "exclusive");
			if (pszParam)
			{
				if (!stricmp(pszParam, "off") || !stricmp(pszParam, "all"))
					PropogateMessage(pszOn, 1);
				else
					PropogateMessage(pszOff, 1);
				g_pMalloc->Free(pszParam);
			}
			else
				PropogateMessage(pszOff, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(0);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 0)
				ToggleLevers(0);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

struct SiblingRelayToggle_Struct
{
	int iState;
	const char* pszMsg;
};

int cScr_SiblingRelay::ToggleIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	cScr_SiblingRelay* pRelay = static_cast<cScr_SiblingRelay*>(pScript);
	SiblingRelayToggle_Struct* pToggleData = reinterpret_cast<SiblingRelayToggle_Struct*>(pData);
	if (pToggleData->iState > 0)
		++pRelay->m_iSwallowTurnOn;
	else if (pToggleData->iState == 0)
		++pRelay->m_iSwallowTurnOff;
	sLink sl;
	pLQ->Link(&sl);
	// Oh hell, let's be greedy.
	SimpleSend(pRelay->m_iObjId, sl.dest, pToggleData->pszMsg);
	return 1;
}

// This pisses me off. How friggen' hard is it to synchronize the state of
// a lever without creating a message loop.
void cScr_SiblingRelay::ToggleLevers(int iState)
{
	SiblingRelayToggle_Struct sToggleData;
	int iLeverState = iState ^ (GetObjectParamInt(m_iObjId, "invert_levers") != 0);
	iLeverState ^= ((m_iFlags & kTrapFlagInvert) == kTrapFlagInvert);
	sToggleData.pszMsg = (iLeverState) ? "TurnOn" : "TurnOff";
	if (GetObjectParamInt(m_iObjId, "no_swallow"))
		sToggleData.iState = -1;
	else
		sToggleData.iState = iState;

#if (_DARKGAME == 3)
	IterateLinksByData("ScriptParams", m_iObjId, 0, "Owns", -1, ToggleIterFunc, this, reinterpret_cast<void*>(&sToggleData));
#else
	IterateLinks("Owns", m_iObjId, 0, ToggleIterFunc, this, reinterpret_cast<void*>(&sToggleData));
#endif
}

// Hell, at least this part worked without incident
void cScr_SiblingRelay::PropogateMessage(const char* pszMsg, int param)
{
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			// I think I'll have the originating object send the message.
			// Rather than going peer-to-peer.
			SimpleSend(m_iObjId, slInfo.dest, pszMsg, param);
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
}

// Traverse the sibling links looking for the lowest object ID.
int cScr_SiblingRelay::FindFirstObject(int iBase)
{
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			if (iBase > slInfo.dest || iBase == 0)
				iBase = slInfo.dest;
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
	return iBase;
}


/******************
 * TrigContact
 */
long __stdcall cScr_ContactTrigger::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->SubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

long __stdcall cScr_ContactTrigger::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->UnsubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

struct ContactTrigger_Struct
{
	bool bOn;
	int iObject;
};

long __stdcall cScr_ContactTrigger::OnPhysContactCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = true;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

long __stdcall cScr_ContactTrigger::OnPhysContactDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = false;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

int cScr_ContactTrigger::FindContactObject(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	if (pContactInfo->bOn)
		pContactTrigger->DoTurnOn(0);
	else
		pContactTrigger->DoTurnOff(0);
	return 0;
}

int cScr_ContactTrigger::FindContactArchetype(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	SService<IObjectSrv> pOS(g_pScriptManager);
	sLink sl;
	pLQ->Link(&sl);
	true_bool child_p;
	if (*(pOS->InheritsFrom(child_p, pContactInfo->iObject, sl.dest)))
	{
		if (pContactInfo->bOn)
			pContactTrigger->DoTurnOn(0);
		else
			pContactTrigger->DoTurnOff(0);
		return 0;
	}
	return 1;
}


#if (_DARKGAME != 3)
/*************
 * TrapText
 */
long cScr_TrapText::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (ShowBook(m_iObjId, false))
		{
			char* psz = GetObjectParams(m_iObjId);
			if (psz)
			{
				ShowString(psz, CalcTextTime(psz), 0);
				g_pMalloc->Free(psz);
			}
		}
	}
	else
	{
		ShowString("", 1);
	}
	return 0;
}

long cScr_TrapText::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrMsg MyMsg;
	MyMsg.message = "TurnOn";
	MyMsg.from = pMsg->from;
	MyMsg.to = pMsg->to;
	MyMsg.time = pMsg->time;
	MyMsg.flags = pMsg->flags;
	return ReceiveMessage(&MyMsg, pReply, kNoAction);
}
#endif


#if (_DARKGAME != 3)
/****************
 * FocusPlaque
 */
long __stdcall cScr_FocusPlaque::OnFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	ShowBook(m_iObjId, false);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long __stdcall cScr_FocusPlaque::OnDeFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if ((static_cast<int>(mpTrapFlags) & kTrapFlagNoOff) == 0)
		{
			ShowString("", 1);
			return 0;
		}
	}
	else
	{
			ShowString("", 1);
			return 0;
	}
	return 1;
}
#endif


/********************
 * ContainerPhysics
 */
long __stdcall cScr_ContainerPhysics::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SInterface<IContainSys> pContainSys(g_pScriptManager);
	sContainIter* pItem = pContainSys->IterStart(m_iObjId);
	if (pItem)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		SService<IObjectSrv> pObjSrv(g_pScriptManager);
		SService<IPhysSrv> pPhysSrv(g_pScriptManager);
		SService<IDataSrv> pDataSrv(g_pScriptManager);
		cScrVec vDims;
		if (pPropSrv->Possessed(m_iObjId, "PhysType"))
		{
			cMultiParm mpShape;
			pPropSrv->Get(mpShape, m_iObjId, "PhysType", "Type");
			switch (static_cast<int>(mpShape))
			{
			case kPMT_Sphere:
			case kPMT_SphereHat:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Radius 1");
				vDims.pos.x = static_cast<float>(mpShape);
				vDims.pos.y = static_cast<float>(mpShape);
				vDims.pos.z = static_cast<float>(mpShape);
				break;
			case kPMT_OBB:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Size");
				vDims.pos.x = static_cast<const mxs_vector*>(mpShape)->x / 2;
				vDims.pos.y = static_cast<const mxs_vector*>(mpShape)->y / 2;
				vDims.pos.z = static_cast<const mxs_vector*>(mpShape)->z / 2;
				mpShape.Free();
				break;
			default:
				break;
			}
		}
		cScrVec vVelocity;
		pPhysSrv->GetVelocity(m_iObjId, vVelocity);
		float fAbsorption = (static_cast<sSlayMsg*>(pMsg)->kind == -900) ? -0.1 : 0.6;
		vVelocity.pos.x *= fAbsorption;
		vVelocity.pos.y *= fAbsorption;
		vVelocity.pos.z *= fAbsorption;
		cScrVec vOffset;
		cScrVec vZero;
		while (1)
		{
			vOffset.pos.x = vDims.pos.x * pDataSrv->RandFltNeg1to1();
			vOffset.pos.y = vDims.pos.y * pDataSrv->RandFltNeg1to1();
			vOffset.pos.z = vDims.pos.z * pDataSrv->RandFltNeg1to1();
			pObjSrv->Teleport(pItem->Object, vOffset, vZero, m_iObjId);
			pContainSys->Remove(pItem->Object, 0);
			pPropSrv->Remove(pItem->Object, "PhysType");
			pPropSrv->Add(pItem->Object, "PhysType");
			pPhysSrv->SetVelocity(pItem->Object, vVelocity);
			if (!pContainSys->IterNext(pItem))
				break;
		}
		pContainSys->IterEnd(pItem);
	}
	return 0;
}


/*******************
 * RandomPatrolPath
 */
long __stdcall cScr_RandomPatrolPath::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);

		std::vector<long> vLinks;
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("AICurrentPatrol"), m_iObjId, 0);
		for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
		{
			vLinks.insert(vLinks.begin(), lsLinks.query->ID());
		}

		if (!vLinks.empty())
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			int r = pDS->RandInt(0, vLinks.size() - 1);
			long dest = vLinks[r];
			vLinks.erase(vLinks.begin() + r);
			std::vector<long>::const_iterator p,q;
			for (p = vLinks.begin(), q = vLinks.end(); p != q; ++p)
			{
				pLS->Destroy(*p);
			}

			if (GetObjectParamInt(m_iObjId, "teleport_ai"))
			{
				sLink sl;
				pLTS->LinkGet(dest, sl);
				SService<IObjectSrv> pOS(g_pScriptManager);
				cScrVec z;
				pOS->Teleport(m_iObjId, z, z, sl.dest);
			}
		}
	}
	return 1;
}


/***********************
 * RandomStartingPoint
 */
long __stdcall cScr_RandomStartingPoint::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IObjectSrv> pObjSrv(g_pScriptManager);
#if (_DARKGAME == 3)
		SService<INetworkingSrv> pNet(g_pScriptManager);
		if (pNet->IsPlayer(m_iObjId))
			return 1;
#else
		object oAvatar;
		pObjSrv->Named(oAvatar, "Avatar");
		true_bool bIsAvatar;
		if (*(pObjSrv->InheritsFrom(bIsAvatar, m_iObjId, oAvatar)))
			return 1;
#endif

		int iDiff = -1;
		{
			SService<IQuestSrv> pQvars(g_pScriptManager);
			if (pQvars->Exists("difficulty"))
			{
				iDiff = pQvars->Get("difficulty");
				if (iDiff > 9) iDiff = 9;
			}
		}

		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);

		std::vector<long> vLinks;
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
		for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
		{
			char* pLinkData = reinterpret_cast<char*>(lsLinks.query->Data());
			if (pLinkData && !strnicmp(pLinkData, "start", 5))
			{
				if (iDiff >= 0)
				{
					if (!strchr(pLinkData, '0' + iDiff))
						continue;
				}
				vLinks.insert(vLinks.begin(), lsLinks.query->ID());
			}
		}

		if (!vLinks.empty())
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			sLink sl;
			pLTS->LinkGet(vLinks.at(pDS->RandInt(0, vLinks.size() - 1)), sl);
			if (sl.dest && sl.dest != m_iObjId)
			{
				cScrVec vZero;
				pObjSrv->Teleport(m_iObjId, vZero, vZero, sl.dest);
			}
		}
	}
	return 1;
}


/***********************
 * TrigDoorOpening
 */
long __stdcall cScr_TrigDoorOpening::OnDoorOpening(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	const char *szMsg = "GoForward";
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & kTrapFlagInvert)
			szMsg = "GoReverse";
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TripFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & 0x10)
			szMsg = "GoReverse";
	}
#else
	if (GetObjectParamInt(m_iObjId, "reverse"))
		szMsg = "GoReverse";
#endif
	CDSend(szMsg, m_iObjId);
	return 0;
}

long __stdcall cScr_TrigDoorOpening::OnDoorClosing(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	const char *szMsg = "GoReverse";
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & kTrapFlagInvert)
			szMsg = "GoForward";
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TripFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & 0x10)
			szMsg = "GoForward";
	}
#else
	if (GetObjectParamInt(m_iObjId, "reverse"))
		szMsg = "GoForward";
#endif
	CDSend(szMsg, m_iObjId);
	return 0;
}

long __stdcall cScr_TrigDoorOpening::OnDoorHalt(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	CDSend("TurnOff", m_iObjId);
	return 0;
}


#if (_DARKGAME != 3)
/***********************
 * InvCarryOver
 */
long __stdcall cScr_InvCarryOver::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (! static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		cAnsiStr sPrefix;
		char* psz = GetObjectParamString(m_iObjId, "inv_code");
		if (psz)
		{
			sPrefix = psz;
			g_pMalloc->Free(psz);
		}
		sPrefix += "inv_";

		SInterface<ITraitManager> pTraits(g_pScriptManager);
		SService<IObjectSrv> pOS(g_pScriptManager);
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		SService<IQuestSrv> pQSrv(g_pScriptManager);

		object oContainer = m_iObjId;
		if (GetObjectParamInt(m_iObjId, "inv_player"))
		{
			pOS->Named(oContainer, "Player");
			if (!oContainer)
				return 1;
		}
		else
		{
			object oPlayer;
			pOS->Named(oPlayer, "Player");
			if (oPlayer == m_iObjId)
			{
				// Don't do anything if this is the Player object.
				// Leave it up to the starting-point to do the work.
				return 1;
			}
			// See if this is the starting-point marker
			// and redirect to the Player if so.
			true_bool __p;
			pLS->AnyExist(__p, pLTS->LinkKindNamed("PlayerFactory"), m_iObjId, 0);
			if (__p)
			{
				if (oPlayer != 0)
					oContainer = oPlayer;
			}
		}

		SService<IDebugScrSrv> pDS(g_pScriptManager);
		cScrStr empty;
		char num[16];
		{
			sprintf(num, " (%d)", static_cast<int>(oContainer));
			cScrStr name;
			pOS->GetName(name, oContainer);
			pDS->MPrint("Carrying Over contents of ", name, num, empty,
						empty, empty, empty, empty);
		}
		
		std::vector<int> vArchetypes;
		{
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
			for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
			{
				char* pLinkData = reinterpret_cast<char*>(lsLinks.query->Data());
				if (pLinkData && !stricmp(pLinkData, "CarryOver"))
				{
					sLink sl;
					lsLinks.query->Link(&sl);
					vArchetypes.insert(vArchetypes.end(), sl.dest);
				}
			}
		}

		if (! vArchetypes.empty())
		{
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("Contains"), oContainer, 0);
			for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
			{
				sLink sl;
				lsLinks.query->Link(&sl);
				for (std::vector<int>::const_iterator arc = vArchetypes.begin(); arc != vArchetypes.end(); ++arc)
				{
					true_bool _p;
					if (*(pOS->InheritsFrom(_p, sl.dest, *arc)))
					{
						object oArc = pTraits->GetArchetype(sl.dest);
						cScrStr sArcName;
						pOS->GetName(sArcName, oArc);
						cAnsiStr sQVar = sPrefix + static_cast<const char*>(sArcName);
						cMultiParm mpCount = 0;
						if (pPropSrv->Possessed(sl.dest, "StackCount"))
							pPropSrv->Get(mpCount, sl.dest, "StackCount", NULL);
						if (pQSrv->Exists(sQVar))
							pQSrv->Delete(sQVar);
						pQSrv->Set(sQVar, mpCount, 1);

						sprintf(num, "%d", static_cast<int>(mpCount));
						pDS->MPrint("  <- ", static_cast<const char*>(sQVar), " = ", num, 
									empty, empty, empty, empty);
						break;
					}
				}
			}
		}
	}
	return 1;
}


/***********************
 * InvLoadUp
 */
long __stdcall cScr_InvLoadUp::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IPropertySrv> pPS(g_pScriptManager);
		if (pPS->Possessed(m_iObjId, "DiffScript"))
		{
			SService<IQuestSrv> pQS(g_pScriptManager);
			cMultiParm mpProp;
			pPS->Get(mpProp, m_iObjId, "DiffScript", NULL);
			if (!(static_cast<int>(mpProp) & (1 << pQS->Get("difficulty"))))
			{
				ClearLoadUp();
			}
		}
		else
		{
			DoLoadUp();
		}
	}
	return 1;
}

long __stdcall cScr_InvLoadUp::OnDifficulty(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoLoadUp();
	return 1;
}

void cScr_InvLoadUp::DoLoadUp(void)
{
	cAnsiStr sPrefix;
	char* psz = GetObjectParamString(m_iObjId, "inv_code");
	if (psz)
	{
		sPrefix = psz;
		g_pMalloc->Free(psz);
	}
	sPrefix += "inv_";

	SInterface<IQuestData> pQuest(g_pScriptManager);
	SService<IContainSrv> pContains(g_pScriptManager);
	SService<IObjectSrv> pOS(g_pScriptManager);
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	
	object oContainer = m_iObjId;
	if (GetObjectParamInt(m_iObjId, "inv_player"))
	{
		pOS->Named(oContainer, "Player");
		if (!oContainer)
		{
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			link pfLink;
			pLS->GetOne(pfLink, pLTS->LinkKindNamed("PlayerFactory"), 0, 0);
			if (pfLink == 0)
				return;
			sLink sl;
			pLTS->LinkGet(pfLink, sl);
			oContainer = sl.source;
		}
	}
	else
	{
		object oPlayer;
		pOS->Named(oPlayer, "Player");
		if (oPlayer == m_iObjId)
		{
			// Don't do anything if this is the Player object.
			// Leave it up to the starting-point to do the work.
			return;
		}
		// See if this is the starting-point marker
		// and redirect to the Player if so.
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool __p;
		pLS->AnyExist(__p, pLTS->LinkKindNamed("PlayerFactory"), m_iObjId, 0);
		if (__p)
		{
			if (oPlayer != 0)
				oContainer = oPlayer;
		}
	}

	SService<IDebugScrSrv> pDS(g_pScriptManager);
	cScrStr empty;
	char num[16];
	{
		sprintf(num, " (%d)", static_cast<int>(oContainer));
		cScrStr name;
		pOS->GetName(name, oContainer);
		pDS->MPrint("Loading Up contents of ", name, num, empty,
					empty, empty, empty, empty);
	}

	IQuestDataIter* qvIter = pQuest->Iter(1);
	for (; !qvIter->Done(); qvIter->Next())
	{
		const char* qvName = qvIter->Name();
		if (!sPrefix.Compare(sPrefix.GetLength(), qvName))
		{
			const char* szArc = qvName + sPrefix.GetLength();
			object oArc, oInv;
			pOS->Named(oArc, szArc);
			if (oArc != 0)
			{
				sprintf(num, "%d", qvIter->Value());
				pDS->MPrint("  -> ", szArc, " = ", num, empty, empty, empty, empty);
				
				pOS->Create(oInv, oArc);
				if (oInv != 0)
				{
					if (qvIter->Value() != 0)
					{
						if (!pPropSrv->Possessed(oInv, "StackCount"))
							pPropSrv->Add(oInv, "StackCount");
						pPropSrv->Set(oInv, "StackCount", qvIter->Value());
					}
					pContains->Add(oInv, oContainer, 0, 1);
				}
			}
			else
			{
				pDS->MPrint("  unknown archetype: ", szArc, empty, empty, empty, empty, empty, empty);
			}
			pQuest->Delete(qvName);
		}
	}
}

void cScr_InvLoadUp::ClearLoadUp(void)
{
	cAnsiStr sPrefix;
	char* psz = GetObjectParamString(m_iObjId, "inv_code");
	if (psz)
	{
		sPrefix = psz;
		g_pMalloc->Free(psz);
	}
	sPrefix += "inv_";

	SInterface<IQuestData> pQuest(g_pScriptManager);
	IQuestDataIter* qvIter = pQuest->Iter(1);
	for (; !qvIter->Done(); qvIter->Next())
	{
		const char* qvName = qvIter->Name();
		if (!sPrefix.Compare(sPrefix.GetLength(), qvName))
		{
			pQuest->Delete(qvName);
		}
	}
}
#endif


/***********************
 * GenericPotion
 */
int cScr_GenericPotion::GetRecipient(void)
{
	int dest = GetObjectParamObject(m_iObjId, "potion_dest");
	if (dest == 0)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), 0, m_iObjId);
		for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
		{
			char* pLinkData = reinterpret_cast<char*>(lsLinks.query->Data());
			if (pLinkData && !stricmp(pLinkData, "Potion"))
			{
				sLink sl;
				lsLinks.query->Link(&sl);
				dest = sl.source;
				break;
			}
		}
	}
	return dest;
}

long __stdcall cScr_GenericPotion::OnStartPotion(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bActive == 0)
	{
		int mp = GetObjectParamObject(m_iObjId, "potion_effect");
		int dest = GetRecipient();
		if (mp != 0 && dest != 0)
		{
			SService<IObjectSrv> pOS(g_pScriptManager);
			pOS->AddMetaProperty(dest, mp);
		}
		m_bActive = 1;
		return 0;
	}
	return 1;
}

long __stdcall cScr_GenericPotion::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_bActive != 0)
	{
		int mp = GetObjectParamObject(m_iObjId, "potion_effect");
		int dest = GetRecipient();
		if (mp != 0 && dest != 0)
		{
			SService<IObjectSrv> pOS(g_pScriptManager);
			pOS->RemoveMetaProperty(dest, mp);
		}
		m_bActive = 0;
		return 0;
	}
	return 1;
}


#if (_DARKGAME != 3)
/***********************
 * InventoryMetaprop
 */
long __stdcall cScr_InvMP::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_bSim)
		return 0;
	int container = static_cast<sContainedScrMsg*>(pMsg)->container;
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oAvatar;
	pOS->Named(oAvatar, "Avatar");
	true_bool bCond;
	if (*(pOS->InheritsFrom(bCond, container, oAvatar)))
	{
		int mp = GetObjectParamObject(m_iObjId, "inv_effect");
		if (mp != 0)
		{
			if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainAdd || 
				static_cast<sContainedScrMsg*>(pMsg)->event == kContainCombine)
			{
				if (!*(pOS->HasMetaProperty(bCond, container, mp)))
				{
					pOS->AddMetaProperty(container, mp);
				}
			}
			else if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainRemove)
			{
				if (*(pOS->HasMetaProperty(bCond, container, mp)))
				{
					pOS->RemoveMetaProperty(container, mp);
				}
			}
		}
		return 0;
	}
	return 1;
}
#endif


/***********************
 * TrapContainer
 */
long __stdcall cScr_ContainerTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

struct ContainerTrap_Struct
{
	IContainSrv* pCont;
	int recipient;
	long create_link;
};

int cScr_ContainerTrap::ContainObjects(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContainerTrap_Struct *cont_info = reinterpret_cast<ContainerTrap_Struct*>(pData);
	cScr_ContainerTrap *pContainerTrap = static_cast<cScr_ContainerTrap*>(pScript);

	sLink sl;
	pLQ->Link(&sl);
	cont_info->pCont->Remove(sl.dest, 0);
	cont_info->pCont->Add(sl.dest, cont_info->recipient, 0, 0);
	if (cont_info->create_link != 0)
	{
		link l;
		pLS->Create(l, cont_info->create_link, pContainerTrap->m_iObjId, sl.dest);
#if (_DARKGAME == 3)
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		pLTS->LinkSetData(l, NULL, "Owns");
#endif
	}
	return 1;
}

long cScr_ContainerTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	int container = GetOneLinkDest(
#if (_DARKGAME == 3)
		"SwitchLink",
#else
		"ControlDevice", 
#endif
		m_iObjId);
	if (GetObjectParamInt(m_iObjId, "track_contains"))
	{
		ContainerTrap_Struct cont_info;
		SService<IContainSrv> pContains(g_pScriptManager);
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cont_info.pCont = pContains;
		if (bTurnOn)
		{
			cont_info.recipient = container;
			cont_info.create_link = 
#if (_DARKGAME == 3)
				pLTS->LinkKindNamed("ScriptParams");
#else
				pLTS->LinkKindNamed("Owns");
#endif
			IterateLinks("Contains", m_iObjId, 0, ContainObjects, this, reinterpret_cast<void*>(&cont_info));
		}
		else
		{
			cont_info.recipient = m_iObjId;
			cont_info.create_link = 0;
#if (_DARKGAME == 3)
			IterateLinksByData("ScriptParams", m_iObjId, 0, "Owns", -1, ContainObjects, this, reinterpret_cast<void*>(&cont_info));
			DestroyAllLinksByData("ScriptParams", m_iObjId, 0, "Owns", -1);
#else
			IterateLinks("Owns", m_iObjId, 0, ContainObjects, this, reinterpret_cast<void*>(&cont_info));
			cMultiParm mpObjId = m_iObjId;
			pLS->DestroyMany(pLTS->LinkKindNamed("Owns"), static_cast<const char*>(mpObjId), "0");
#endif
		}
	}
	else
	{
		if (container != 0)
		{
			SService<IContainSrv> pContain(g_pScriptManager);
			if (bTurnOn)
			{
				pContain->MoveAllContents(m_iObjId, container, 1);
			}
			else
			{
				pContain->MoveAllContents(container, m_iObjId, 1);
			}
		}
	}
#if (_DARKGAME == 3)
	SService<IShockGameSrv> pShock(g_pScriptManager);
	pShock->RefreshInv();
#endif
	return 0;
}


/***********************
 * TrapDispenser
 */
long __stdcall cScr_DispenserTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

long cScr_DispenserTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (!bTurnOn)
		return 0;

	int container = GetOneLinkDest(
#if (_DARKGAME == 3)
		"SwitchLink",
#else
		"ControlDevice", 
#endif
		m_iObjId);
	if (container != 0)
	{
		int iArc = GetObjectParamObject(m_iObjId, "dispense");
		if (iArc == 0)
			return 0;

		SService<IObjectSrv> pOS(g_pScriptManager);
		object oEmit;
		pOS->Create(oEmit, iArc);
		if (oEmit)
		{
			SService<IContainSrv> pContains(g_pScriptManager);
			pContains->Add(oEmit, container, 0, 1);
#if (_DARKGAME == 3)
			SService<IShockGameSrv> pShock(g_pScriptManager);
			pShock->RefreshInv();
#endif
		}
	}
	return 0;
}


#if (_DARKGAME != 3)
/***********************
 * TrigPurchase
 */
long __stdcall cScr_PurchaseTrigger::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "DidHeBuyIt", 100, kSTM_OneShot, 0);
	}
	return 1;
}

long __stdcall cScr_PurchaseTrigger::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "DidHeBuyIt"))
	{
		InitTrigVars();
		m_iFlags &= (kTrapFlagNoOn | kTrapFlagNoOff | kTrapFlagInvert);
		object oAvatar = StrToObject("Player");
		if (oAvatar == 0)
			return 0;
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool bCond;
		if (*(pLS->AnyExist(bCond, pLTS->LinkKindNamed("Contains"), oAvatar, cScript::m_iObjId)))
		{
			DoTurnOn(0);
		}
		else
		{
			DoTurnOff(0);
		}
		return 0;
	}
	return 1;
}
#endif


#if (_DARKGAME == 2)
/***********************
 * SimpleGun
 */
long __stdcall cScr_Gun::OnSelect(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IWeaponSrv> pWeapon(g_pScriptManager);
	pWeapon->Equip(m_iObjId, 0);
	return 0;
}

long __stdcall cScr_Gun::OnDeSelect(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IWeaponSrv> pWeapon(g_pScriptManager);
	pWeapon->UnEquip(m_iObjId);
	return 0;
}

long __stdcall cScr_Gun::OnFrobBegin(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sLink sl;
	link l = GetAnyLinkInheritedSrc("Projectile", m_iObjId, 0, &sl);
	if (l != 0)
	{
		object oProj = sl.dest;
		SService<IPhysSrv> pPhys(g_pScriptManager);
		cScrVec vecVel;
		object oEmit;
		pPhys->LaunchProjectile(oEmit, m_iObjId, oProj, 1.0, 6, vecVel);

		l = GetAnyLinkInheritedSrc("GunFlash", m_iObjId, 0, &sl);
		if (l != 0)
		{
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			SService<IObjectSrv> pOS(g_pScriptManager);
			object oFlash = 0;
			pOS->BeginCreate(oFlash, sl.dest);
			if (oFlash)
			{
				object oAttach;
				int vhot;
				cMultiParm mpFlashInfo;
				pLTS->LinkGetData(mpFlashInfo, l, "VHot");
				vhot = static_cast<int>(mpFlashInfo);
				pLTS->LinkGetData(mpFlashInfo, l, "Flags");
				pPhys->SetGravity(oFlash, 0);
				if (static_cast<int>(mpFlashInfo) & 1)
					oAttach = oEmit;
				else
					oAttach = m_iObjId;
				pOS->Teleport(oFlash, cScrVec(), cScrVec(), oAttach);
				l = 0;
				pLS->Create(l, pLTS->LinkKindNamed("ParticleAttachement"), oFlash, oAttach);
				if (l)
				{
					pLTS->LinkSetData(l, "Type", 1);
					pLTS->LinkSetData(l, "vhot #", vhot);
				}
				pOS->SetTransience(oFlash, true);
				pOS->EndCreate(oFlash);
			}
		}
	}
	return 0;
}

long __stdcall cScr_Gun::OnFrobEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	return 1;
}
#endif


/***********************
 * TurnAtWaypoints
 */
bool cScr_TurnVator::AdjustTweq(const mxs_vector& currvec, const mxs_vector& destvec)
{
	bool activate = false;
	mxs_vector cfg;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	const mxs_vector* rate = m_vTweqRate;

	// X-axis
	if (rate->x > 0.0 && fabs(destvec.x - currvec.x) > 0.005)
	{
		double delta = destvec.x - currvec.x;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.x > currvec.x)
			{
				if (currvec.x < 0.005)
				{
					cfg.x = rate->x;
					cfg.y = destvec.x;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Bank", 0xFFFF);
				}
				else
				{
					cfg.x = rate->x;
					cfg.y = 0.0f;
					cfg.z = currvec.x;
				}
			}
			else
			{
				cfg.x = rate->x;
				cfg.y = destvec.x;
				cfg.z = currvec.x;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 3);
		}
		else
		{
			if (destvec.x < currvec.x)
			{
				if (currvec.x > 359.99)
				{
					cfg.x = rate->x;
					cfg.y = 0.0f;
					cfg.z = destvec.x;
					pPropSrv->Set(m_iObjId, "Position", "Bank", 0);
				}
				else
				{
					cfg.x = rate->x;
					cfg.y = currvec.x;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->x;
				cfg.y = currvec.x;
				cfg.z = destvec.x;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 1);
		}
		activate = true;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", &cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 0);
	}

	// Y-axis
	if (rate->y > 0.0 && fabs(destvec.y - currvec.y) > 0.005)
	{
		double delta = destvec.y - currvec.y;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.y > currvec.y)
			{
				if (currvec.y < 0.005)
				{
					cfg.x = rate->y;
					cfg.y = destvec.y;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Pitch", 0xFFFF);
				}
				else
				{
					cfg.x = rate->y;
					cfg.y = 0.0f;
					cfg.z = currvec.y;
				}
			}
			else
			{
				cfg.x = rate->y;
				cfg.y = destvec.y;
				cfg.z = currvec.y;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 3);
		}
		else
		{
			if (destvec.y < currvec.y)
			{
				if (currvec.y > 359.99)
				{
					cfg.x = rate->y;
					cfg.y = 0.0f;
					cfg.z = destvec.y;
					pPropSrv->Set(m_iObjId, "Position", "Pitch", 0);
				}
				else
				{
					cfg.x = rate->y;
					cfg.y = currvec.y;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->y;
				cfg.y = currvec.y;
				cfg.z = destvec.y;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 1);
		}
		activate = true;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", &cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 0);
	}

	// Z-axis
	if (rate->z > 0.0 && fabs(destvec.z - currvec.z) > 0.005)
	{
		double delta = destvec.z - currvec.z;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.z > currvec.z)
			{
				if (currvec.z < 0.005)
				{
					cfg.x = rate->z;
					cfg.y = destvec.z;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Heading", 0xFFFF);
				}
				else
				{
					cfg.x = rate->z;
					cfg.y = 0.0f;
					cfg.z = currvec.z;
				}
			}
			else
			{
				cfg.x = rate->z;
				cfg.y = destvec.z;
				cfg.z = currvec.z;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 3);
		}
		else
		{
			if (destvec.z < currvec.z)
			{
				if (currvec.z > 359.99)
				{
					cfg.x = rate->z;
					cfg.y = 0.0f;
					cfg.z = destvec.z;
					pPropSrv->Set(m_iObjId, "Position", "Heading", 0);
				}
				else
				{
					cfg.x = rate->z;
					cfg.y = currvec.z;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->z;
				cfg.y = currvec.z;
				cfg.z = destvec.z;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", &cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 1);
		}
		activate = true;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", &cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 0);
	}

	cMultiParm mpAnims = (activate) ? 1 : 0;
	pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", mpAnims);
	return activate;
}

long __stdcall cScr_TurnVator::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	mxs_vector vRate = {0.0, 0.0, 0.0};
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		cMultiParm mpRate;
		pPropSrv->Get(mpRate, m_iObjId, "CfgTweqRotate", "x rate-low-high");
		vRate.x = static_cast<const mxs_vector*>(mpRate)->x;
		mpRate.Free();
		pPropSrv->Get(mpRate, m_iObjId, "CfgTweqRotate", "y rate-low-high");
		vRate.y = static_cast<const mxs_vector*>(mpRate)->x;
		mpRate.Free();
		pPropSrv->Get(mpRate, m_iObjId, "CfgTweqRotate", "z rate-low-high");
		vRate.z = static_cast<const mxs_vector*>(mpRate)->x;
		mpRate.Free();

		cMultiParm mpCfg;
		pPropSrv->Get(mpCfg, m_iObjId, "CfgTweqRotate", "Primary Axis");
		if (static_cast<int>(mpCfg) != 0)
		{
			mpCfg = 0;
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "Primary Axis", mpCfg);
		}
	}
	m_vTweqRate.Init(&vRate);
	return 0;
}

long __stdcall cScr_TurnVator::OnStart(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		int x,y,z;
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 1AnimS");
		x = static_cast<int>(mpFlags);
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 2AnimS");
		y = static_cast<int>(mpFlags);
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 3AnimS");
		z = static_cast<int>(mpFlags);
		if (x != 0 || y != 0 || z != 0)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", mpFlags);
		}
	}
	return 0;
}

long __stdcall cScr_TurnVator::OnStop(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		cMultiParm mpFlags = 0;
		pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", mpFlags);
	}
	return 0;
}

long __stdcall cScr_TurnVator::OnWaypoint(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		cMultiParm mpFlags = 0;
		pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", mpFlags);

		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec currvec, destvec;
		pOS->Facing(currvec, m_iObjId);
		pOS->Facing(destvec, static_cast<sMovingTerrainMsg*>(pMsg)->waypoint);
		if (AdjustTweq(currvec.pos, destvec.pos))
		{
			sMultiParm mpDestVec;
			mpDestVec.type = kMT_Vector;
			mpDestVec.pVector = &destvec.pos;
			sScrDatumTag tagDestVec;
			tagDestVec.objId = m_iObjId;
			tagDestVec.pszClass = "cScr_TurnVator";
			tagDestVec.pszName = "m_vDestVec";
			g_pScriptManager->SetScriptData(&tagDestVec, &mpDestVec);
		}
	}
	return 0;
}

long __stdcall cScr_TurnVator::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	if (pTweqMsg->Type == kTweqTypeRotate && pTweqMsg->Op == kTweqOpHaltTweq)
	{
		sScrDatumTag tagDestVec;
		tagDestVec.objId = m_iObjId;
		tagDestVec.pszClass = "cScr_TurnVator";
		tagDestVec.pszName = "m_vDestVec";
		if (g_pScriptManager->IsScriptDataSet(&tagDestVec))
		{
			sMultiParm mpDestVec;
			mpDestVec.type = kMT_Undef;
			g_pScriptManager->GetScriptData(&tagDestVec, &mpDestVec);

			SService<IObjectSrv> pOS(g_pScriptManager);
			cScrVec currvec;
			pOS->Facing(currvec, m_iObjId);
			if (! AdjustTweq(currvec.pos, *mpDestVec.pVector))
			{
				sMultiParm mpRet;
				mpRet.type = kMT_Undef;
				g_pScriptManager->ClearScriptData(&tagDestVec, &mpRet);
				g_pMalloc->Free(mpRet.pVector);
			}
			g_pMalloc->Free(mpDestVec.pVector);
		}
	}
	return 0;
}


/***********************
 * VisibleRelay
 */
bool cScr_VisibleTrap::TestVisible(bool bVisible)
{
	int iObjId = m_iObjId;
	sLink slVisLink;
	if (GetOneLinkByData("ScriptParams", m_iObjId, 0, &slVisLink, "Visible", -1))
	{
		iObjId = slVisLink.dest;
	}
	true_bool bIsVisible;
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	pObjSrv->RenderedThisFrame(bIsVisible, iObjId);
	DebugPrintf("Object %d is %svisible", iObjId, (bIsVisible)?"":"not ");
	return bVisible == bIsVisible;
}

long __stdcall cScr_VisibleTrap::OnQuery(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!pReply)
		return 1;

	const char* pszParm = static_cast<const char*>(pMsg->data);
	if (!pszParm || !*pszParm)
		return 1;

	bool bVisible = true;
	char* pszEnd;
	long iConv = strtol(pszParm, &pszEnd, 10);
	if (pszEnd > pMsg->data.psz)
	{
		switch (iConv)
		{
		case 0:
			bVisible = false; break;
		case 1:
			bVisible = true; break;
		default:
			return 1;
		}
	}
	else
	{
		if (!stricmp(pMsg->data.psz, "visible") || 
			!stricmp(pMsg->data.psz, "on") ||
			!stricmp(pMsg->data.psz, "true") ||
			!stricmp(pMsg->data.psz, "yes"))
		{
			bVisible = true;
		}
		else if (!stricmp(pMsg->data.psz, "invisible") || 
			!stricmp(pMsg->data.psz, "off") ||
			!stricmp(pMsg->data.psz, "false") ||
			!stricmp(pMsg->data.psz, "no"))
		{
			bVisible = false;
		}
		else
			return 1;
	}

	MultiParmAssignInt(pReply, (TestVisible(bVisible)) ? 1 : 0);
	return 0;
}

long __stdcall cScr_VisibleTrap::OnTrigger(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	int iFlags = 0;
#if (_DARKGAME == 2)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}
#elif (_DARKGAME == 3)
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TripFlags"))
	{
		cMultiParm mpTrapFlags = 0;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TripFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
		iFlags = ((iFlags & 0x18) >> 3) | ((iFlags & 0x2) << 1) | ((iFlags & 0x1) << 3);
	}
#else
	if (GetObjectParamInt(m_iObjId, "invisible"))
		iFlags = kTrapFlagInvert;
#endif

	if (TestVisible((iFlags & kTrapFlagInvert) == 0))
	{
		CDSend(pMsg->message, m_iObjId, pMsg->data);
	}
	
#if (_DARKGAME != 1)
	if (iFlags & kTrapFlagOnce)
	{
		cMultiParm mpLock = true;
		pPropSrv->Set(m_iObjId, "Locked", mpLock);
	}
#endif

	return 0;
}


/***********************
 * ResetTweqState
 */
void cScr_ResetTweqs::ResetTweq(const char* pszName, int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, pszName) && pPS->Possessed(iArc, pszName))
	{
		cMultiParm mpTime, mpFrame;
		pPS->Get(mpTime, iArc, pszName, "Cur Time");
		pPS->Get(mpFrame, iArc, pszName, "Frame #");
		pPS->Set(iObj, pszName, "Cur Time", mpTime);
		pPS->Set(iObj, pszName, "Frame #", mpFrame);
	}
}

void cScr_ResetTweqs::ResetProp(const char* pszName, int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, pszName) && pPS->Possessed(iArc, pszName))
	{
		pPS->Remove(iObj, pszName);
		pPS->Add(iObj, pszName);
	}
}

void cScr_ResetTweqs::ResetAll(int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, "StTweqJoints"))
		ResetProp("JointPos", iObj, iArc, pPS);
	ResetTweq("StTweqBlink", iObj, iArc, pPS);
	ResetTweq("StTweqEmit", iObj, iArc, pPS);
	ResetTweq("StTweq2Emit", iObj, iArc, pPS);
	ResetTweq("StTweq3Emit", iObj, iArc, pPS);
	ResetTweq("StTweq4Emit", iObj, iArc, pPS);
	ResetTweq("StTweq5Emit", iObj, iArc, pPS);
	if (pPS->Possessed(iObj, "StTweqModels"))
		ResetProp("ModelName", iObj, iArc, pPS);
	if (pPS->Possessed(iObj, "StTweqScale"))
		ResetProp("Scale", iObj, iArc, pPS);
}

long __stdcall cScr_ResetTweqs::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	SService<IPropertySrv> pPS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	int iArc = pTM->GetArchetype(m_iObjId);

	if (pTweqMsg->Op == kTweqOpHaltTweq)
	{
		switch (pTweqMsg->Type)
		{
		case kTweqTypeJoints:
			ResetProp("JointPos", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeEmitter:
			ResetTweq("StTweqEmit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq2Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq3Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq4Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq5Emit", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeFlicker:
			ResetTweq("StTweqBlink", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeModels:
			ResetProp("ModelName", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeScale:
			ResetProp("Scale", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeAll:
			ResetAll(m_iObjId, iArc, pPS);
		case kTweqTypeRotate:
			if (pPS->Possessed(m_iObjId, "StTweqRotate"))
			{
				SService<IObjectSrv> pOS(g_pScriptManager);
				cScrVec vFacing = *m_vRotInit;
				cScrVec vPos;
				pOS->Position(vPos, m_iObjId);
				pOS->Teleport(m_iObjId, vPos, vFacing, 0);
			}
			break;
		case kTweqTypeDelete:
		case kTweqTypeLock:
		default:
			break;
		}
	}

	return 0;
}

long __stdcall cScr_ResetTweqs::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	int iArc = pTM->GetArchetype(m_iObjId);
	ResetAll(m_iObjId, iArc, pPS);
	if (pPS->Possessed(m_iObjId, "StTweqRotate"))
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vFacing = *m_vRotInit;
		cScrVec vPos;
		pOS->Position(vPos, m_iObjId);
		pOS->Teleport(m_iObjId, vPos, vFacing, 0);
	}
	return 0;
}

long __stdcall cScr_ResetTweqs::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vFacing;
		pOS->Facing(vFacing, m_iObjId);
		m_vRotInit.Init(&vFacing.pos);
	}
	return 1;
}


/***********************
 * RecallDevice
 */
int cScr_RecallDevice::RecallIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iAvatar = reinterpret_cast<int>(pData);
	SService<IObjectSrv> pOS(g_pScriptManager);
	sLink sl;
	pLQ->Link(&sl);
	cScrVec vNil;
	pOS->Teleport(sl.dest, vNil, vNil, iAvatar);
	return 1;
}

long cScr_RecallDevice::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		int iAvatar = StrToObject("Player");
		IterateLinks(
#if (_DARKGAME == 3)
			"SwitchLink", 
#else
			"ControlDevice", 
#endif
			m_iObjId, 0, RecallIter, this, reinterpret_cast<void*>(iAvatar));
	}
	return 0;
}


#if (_DARKGAME != 3)
/***********************
 * TrigProxFrob
 */
long __stdcall cScr_ProxFrobTrigger::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (IsLocked(pMsg))
		return 1;
	int iFrobber = static_cast<sFrobMsg*>(pMsg)->Frobber;
	if (!iFrobber)
		return 1;
	int iProx = GetOneLinkDest("Route", cScript::m_iObjId);
	if (!iProx)
		return 1;
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	true_bool bNearby;
	pLS->AnyExist(bNearby, pLTS->LinkKindNamed("Population"), iProx, iFrobber);
	if (bNearby)
	{
		InitTrigVars();
		DoTurnOn(0);
		SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
		eEnvSoundLoc opt = (iFrobber == StrToObject("Player")) ? kEnvSoundAmbient : kEnvSoundOnObj;
		true_bool _p;
		pSoundSrv->PlayEnvSchema(_p, cScript::m_iObjId, "Event Activate", cScript::m_iObjId, 0, opt SOUND_NET);
	}

	return 0;
}
#endif


